import hmac
from hashlib import sha256
import base64
import struct
from zlib import crc32
import secrets
import time
from collections import OrderedDict

kJoinChannel = 1
kPublishAudioStream = 2
kPublishVideoStream = 3
kPublishDataStream = 4
kRtmLogin = 1000

VERSION_LENGTH = 3
APP_ID_LENGTH = 32


def getVersion():
    return '006'


def packUint16(x):
    return struct.pack('<H', int(x))


def packUint32(x):
    return struct.pack('<I', int(x))


def packInt32(x):
    return struct.pack('<i', int(x))


def packString(string):
    return packUint16(len(string)) + string


def packMap(m):
    ret = packUint16(len(list(m.items())))
    for k, v in list(m.items()):
        ret += packUint16(k) + packString(v)
    return ret


def packMapUint32(m):
    ret = packUint16(len(list(m.items())))
    for k, v in list(m.items()):
        ret += packUint16(k) + packUint32(v)
    return ret


class ReadByteBuffer:

    def __init__(self, bytes):
        self.buffer = bytes
        self.position = 0

    def unPackUint16(self):
        len = struct.calcsize('H')
        buff = self.buffer[self.position: self.position + len]
        ret = struct.unpack('<H', buff)[0]
        self.position += len
        return ret

    def unPackUint32(self):
        len = struct.calcsize('I')
        buff = self.buffer[self.position: self.position + len]
        ret = struct.unpack('<I', buff)[0]
        self.position += len
        return ret

    def unPackString(self):
        strlen = self.unPackUint16()
        buff = self.buffer[self.position: self.position + strlen]
        ret = struct.unpack('<' + str(strlen) + 's', buff)[0]
        self.position += strlen
        return ret

    def unPackMapUint32(self):
        messages = {}
        maplen = self.unPackUint16()

        for index in range(maplen):
            key = self.unPackUint16()
            value = self.unPackUint32()
            messages[key] = value
        return messages


def unPackContent(buff):
    readbuf = ReadByteBuffer(buff)
    signature = readbuf.unPackString()
    crc_channel_name = readbuf.unPackUint32()
    crc_uid = readbuf.unPackUint32()
    m = readbuf.unPackString()

    return signature, crc_channel_name, crc_uid, m


def unPackMessages(buff):
    readbuf = ReadByteBuffer(buff)
    salt = readbuf.unPackUint32()
    ts = readbuf.unPackUint32()
    messages = readbuf.unPackMapUint32()

    return salt, ts, messages


class AccessToken:

    def __init__(self, appID='', appCertificate='', channelName='', uid=''):
        self.appID = appID
        self.appCertificate = appCertificate
        self.channelName = channelName
        self.ts = int(time.time()) + 24 * 3600
        self.salt = secrets.SystemRandom().randint(1, 99999999)
        self.messages = {}
        if (uid == 0):
            self.uidStr = ""
        else:
            self.uidStr = str(uid)

    def addPrivilege(self, privilege, expireTimestamp):
        self.messages[privilege] = expireTimestamp

    def fromString(self, originToken):
        try:
            dk6version = getVersion()
            originVersion = originToken[:VERSION_LENGTH]
            if (originVersion != dk6version):
                return False

            originAppID = originToken[VERSION_LENGTH:(VERSION_LENGTH + APP_ID_LENGTH)]
            originContent = originToken[(VERSION_LENGTH + APP_ID_LENGTH):]
            originContentDecoded = base64.b64decode(originContent)

            signature, crc_channel_name, crc_uid, m = unPackContent(originContentDecoded)
            self.salt, self.ts, self.messages = unPackMessages(m)

        except Exception as e:
            print("error:", str(e))
            return False

        return True

    def build(self):

        self.messages = OrderedDict(sorted(iter(self.messages.items()), key=lambda x: int(x[0])))

        m = packUint32(self.salt) + packUint32(self.ts) \
            + packMapUint32(self.messages)

        val = self.appID.encode('utf-8') + self.channelName.encode('utf-8') + self.uidStr.encode('utf-8') + m

        signature = hmac.new(self.appCertificate.encode('utf-8'), val, sha256).digest()
        crc_channel_name = crc32(self.channelName.encode('utf-8')) & 0xffffffff
        crc_uid = crc32(self.uidStr.encode('utf-8')) & 0xffffffff

        content = packString(signature) \
                  + packUint32(crc_channel_name) \
                  + packUint32(crc_uid) \
                  + packString(m)

        version = getVersion()
        ret = version + self.appID + base64.b64encode(content).decode('utf-8')
        return ret


import os
import sys
import json
import subprocess

from flask import current_app as app
from .validation_errors import JavascriptError, ValidationError
import zmq
context = zmq.Context()

class AjvValidator:

    def __init__(self):
        #  Socket to talk to server
        self.socket = context.socket(zmq.REQ)
        self.socket.setsockopt(zmq.RCVTIMEO, 1000)
        # bind to a random port in loopback iface
        port = self.socket.bind_to_random_port("tcp://127.0.0.1")
        validator_path = os.path.join(app.config['COVFEE_SHARED_PATH'], 'validator', 'validator_service.js')

        # start the nodejs validation server in the same interface
        self.process = subprocess.Popen(["node", '--trace-warnings', validator_path, "serve", str(port)], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        # check that the service is up and running
        if 'Waiting' not in self.process.stdout.readline().decode():
            stderr = self.process.stderr.read().decode()
            e = Exception(f'Error running JS validator {validator_path}')
            e.js_stack_trace = stderr
            raise e
        
    def __del__(self):
        # pass
        self.process.kill()

    @staticmethod
    def parse_ajv_path(ajv_instance_path):
        if ajv_instance_path == '/': return ''

        parts = ajv_instance_path.split('/')[1:]
        return parts

    def get_friendly_error_message(self, err):
        if 'keyword' not in err: return err['message']
        
        if err['keyword'] == 'additionalProperties':
            return err['message'] + f'. Property \'{err["params"]["additionalProperty"]}\' is unrecognized.'

        if err['keyword'] == 'discriminator':
            if err['params']['error'] == 'mapping':
                return f'Invalid value \'{err["params"]["tagValue"]}\' for property \'{err["params"]["tag"]}\'. Please make sure you are using a supported value.'
        return err['message']

    def schema_validate(self, schema_name, data):
        self.socket.send_json({
            'schema': schema_name, 
            'data': data
        })
        message = self.socket.recv()
        result = json.loads(message.decode('utf-8'))
        if result['jsError']:
            e = JavascriptError(f'JS code produced an error.')
            e.js_stack_trace = result['stackTrace']
            raise e
        if not result['valid']:
            err = result['errors'][0]
            raise ValidationError(
                message=self.get_friendly_error_message(err),
                path=AjvValidator.parse_ajv_path(err['instancePath']),
                instance=err['data']
            )

    def validate_project(self, project_spec):
        self.schema_validate('ProjectSpec', project_spec)



from .db import db
from hashlib import pbkdf2_hmac

from flask import current_app as app

def password_hash(password: str):
    return pbkdf2_hmac('sha256', password.encode(), app.config['JWT_SECRET_KEY'].encode(), 10000)

class User(db.Model):
    """ Represents a covfee user """
    __tablename__ = 'users'

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String) # note username is non-unique
    roles = db.Column(db.JSON)

    providers = db.relationship("AuthProvider", backref="user", cascade="all, delete")

    def __init__(self, username: str, roles: list = ['user', 'admin']):
        self.username = username
        self.roles = roles

    def add_provider(self, *args, **kwargs):
        self.providers.append(AuthProvider(*args, **kwargs))

class AuthProvider(db.Model):
    __tablename__ = 'auth_providers'

    provider_id = db.Column(db.String, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey('users.id'), primary_key=True)

    # holds provider specific information like the password for password provider
    extra = db.Column(db.JSON)

    def __init__(self, provider_id: str, user_id: str, extra=None):
        self.provider_id = provider_id
        self.user_id = user_id
        self.extra = extra

    



import os
import json

from flask import current_app as app
from flask import Flask, Blueprint, render_template, send_from_directory
from flask_cors import CORS
from flask_jwt_extended import JWTManager
from flask_socketio import SocketIO

from .orm import db, load_config
from .rest_api import api, auth, admin_required, add_claims_to_access_token, user_identity_lookup, \
    user_loader_callback

def create_app(mode):
    app = Flask(__name__, static_folder=None)
    load_config(app, mode=mode)
    db.init_app(app)
    socketio = SocketIO()
    socketio.init_app(app)

    app.register_blueprint(frontend, url_prefix='/')
    app.register_blueprint(api, url_prefix='/api')
    app.register_blueprint(auth, url_prefix='/auth')
    CORS(app, resources={r"/*": {"origins": "*"}})
    jwt = JWTManager(app)

    jwt.user_claims_loader(add_claims_to_access_token)
    jwt.user_identity_loader(user_identity_lookup)
    jwt.user_loader_callback_loader(user_loader_callback)

    @app.teardown_request
    def teardown_request(exception):
        if exception:
            db.session.rollback()
        db.session.remove()
    
    return socketio, app


# APP ROUTES
frontend = Blueprint('frontend', __name__,
                     template_folder=os.path.join(os.path.dirname(os.path.realpath(__file__)), 'templates'))


# annotation app
@frontend.route('/')
def main():
    return render_template('app.html',
                           constants=json.dumps(app.config['FRONTEND_CONFIG']),
                           bundle_url=app.config['BUNDLE_URL'])


# admin interface
@frontend.route('/admin')
def admin():
    return render_template('admin.html',
                           constants=json.dumps(app.config['FRONTEND_CONFIG']),
                           bundle_url=app.config['BUNDLE_URL'])


# project www server
@frontend.route('/www/<path:filename>')
def project_www_file(filename):
    return send_from_directory(app.config['PROJECT_WWW_PATH'], filename)


@frontend.errorhandler(404)
def page_not_found(e):
    # note that we set the 404 status explicitly
    return render_template('404.html'), 404



import pandas as pd
import numpy as np
from typing import Tuple, List, Any
# from ..orm.task import Task, TaskResponse

class BaseCovfeeTask:

    def __init__(self, response=None, task=None):
        self.response = response

        if response:
            self.task = response.task
        else:
            self.task = task

    def get_task_specific_props(self) -> dict:
        """Used to extend the dict that is send to the browser as props for the task element.

        Returns:
            dict: new props to be merged with the default props
        """
        return {}

    def to_dict(self, with_chunk_data: bool) -> dict:
        """Processes a task response to return a friendly dict (JSON) with task results and metadata
        Called when data is downloaded in JSON format.
        Does not affect the data in the database.

        Args:
            with_chunk_data (bool): If true, include the continuous response data.

        Returns:
            dict: task response and summary metadata
        """
        if with_chunk_data:
            chunk_data, chunk_logs = self.response.get_ndarray()
        else:
            chunk_data, chunk_logs = (None, None)

        return {
            'response': self.response.data,
            'data': chunk_data.tolist() if chunk_data is not None else [],
            'logs': chunk_logs if chunk_logs is not None else [],
            'created_at': self.response.created_at.isoformat(),
            'updated_at': self.response.updated_at.isoformat(),
            'submitted_at': self.response.submitted_at.isoformat(),
        }

    def to_dataframe(self, data: np.ndarray) -> pd.DataFrame:
        """Transforms the submitted data into a pandas dataframe.
        This method is called when results are downloaded as CSV from the admin panel.
        It can be implemented to add headers to the default dataframe or to filter data before download.
        Does not affect the data in the database.

        Args:
            data (bytes): Binary, aggregated data.
        """
        if data is None:
            return None
        else:
            assert data.ndim == 2
            assert data.shape[1] >= 3
            num_columns = data.shape[1] - 2
            return pd.DataFrame(data, columns=['index', 'media_time', *[f'data{i}' for i in range(num_columns)]])

    def validate(self, response: Any, data: np.ndarray = None, log_data: List[List[Any]] = None):
        """This method decides whether a particular task submission will be accepted or not

        Args:
            response (object): The (non-continuous) task response. Usually includes the value of 
                input forms / sliders / radio buttons in the task.
            data (pd.DataFrame): The continuous task response, includes all the data points 
                logged by the task. For continuous tasks this may be the only response.
            log_data (list[list[any]], optional): The logs submitted by the task via buffer.log(). 
                Usually contains auxiliary information. Defaults to None.

        Returns:
            [type]: [description]
        """
        return True

    def on_first_join():
        """ Called when the first visitor joins the task.
            (for socketio-enabled tasks)
        """
        pass

    def on_last_leave():
        """ Called when the last person left leaves the task page
            (for socketio-enabled tasks)
        """
        pass




from construct import Struct, Int32ul, Bytes, this
# const packetParser = new Parser()
#     .endianess("little")
#     .uint32("dataLength")
#     .uint32("logsLength")
#     .uint32("chunkLength")
#     .uint32("recordLength")
#     .buffer("data", {length: "dataLength", clone: true})
#     .string("logs", {length: "logsLength", clone: true})


class Packer:

    def __init__(self):
        self.packetStruct = Struct(
            'dataLength' / Int32ul,
            'logsLength' / Int32ul,
            'chunkLength' / Int32ul,
            'recordLength' / Int32ul,
            'data' / Bytes(this.dataLength),
            'logs' / Bytes(this.logsLength)
        )

    def parseChunk(self, chunk):
        parsed = self.packetStruct.parse(chunk)
        return parsed



import hashlib
import hmac

from ..orm import db
from .. import tasks
from ..tasks.base import BaseCovfeeTask
from flask import current_app as app, session
from flask_socketio import SocketIO, send, emit, join_room, leave_room
from covfee.server.orm.task import TaskResponse
from covfee.server.socketio.redux_store import ReduxStoreClient

socketio = SocketIO()
store = ReduxStoreClient()

def get_task_object(responseId: int):
    response = db.session.query(TaskResponse).get(responseId)
    if response is None:
        return None
    
    task = response.task
    task_class = getattr(tasks, task.spec.spec['type'], BaseCovfeeTask)
    task_object = task_class(response=response)
    return task_object


@socketio.on('message')
def handle_message(data):
    pass


@socketio.on('join')
def on_join(data):
    room = str(data['room'])

    res = store.join(room)
    if res['success']:
        join_room(room)
        session['room'] = room
        emit('state', res)

        # if this is the first join, run the on_first_join callback
        if res['numConnections'] == 1:
            get_task_object(int(room)).on_first_join()

    else:
        send(f'Unable to join room {room}')


@socketio.on('action')
def on_action(data):
    action = data['action']
    room = str(data['room'])
    if room != session['room']:
        return send(f'data["room"] does not match session\'s room variable. {room} != {session["room"]}')

    res = store.action(room, action)
    if res['success']:
        emit('action', res, to=room)


@socketio.on('state')
def on_state(data):
    room = str(data['room'])
    if room != session['room']:
        return send(f'data["room"] does not match session\'s room variable. {room} != {session["room"]}')

    res = store.state(room)
    if res['success']:
        emit('state', res)


@socketio.on('leave')
def on_leave(data):
    room = str(data['room'])
    if room != session['room']:
        return send(f'data["room"] does not match session\'s room variable. {room} != {session["room"]}')

    res = store.leave(room)
    session['room'] = None
    leave_room(room)

    # if the room is now empty
    if res['numConnections'] == 0:
        get_task_object(int(room)).on_last_leave()

@socketio.on('disconnect')
def disconnect():
    if 'room' in session:
        room = session['room']
        res = store.leave(room)
        session['room'] = None
        leave_room(room)






""" Launch commands

These commands launch or build covfee and are supported for the typical use case.
"""
import os
import sys
from colorama import init as colorama_init, Fore
import click
import traceback
from getpass import getpass

from flask import current_app as app
from flask.cli import FlaskGroup, pass_script_info

from covfee.server.start import create_app
from ...covfee_folder import CovfeeFolder, ProjectExistsException
from halo.halo import Halo
from covfee.shared.validator.validation_errors import JavascriptError, ValidationError
from covfee.shared.schemata import Schemata
from covfee.cli.utils import NPMPackage
from covfee import _make as covfee_make, start_deepstream
from covfee.server.rtstore import rtstore
from covfee.server.orm import set_frontend_config

colorama_init()


@click.group(name='covfee')
def covfee_cli():
    pass

def start_covfee(socketio, app, mode='local', host='localhost'):
    if app.config['SSL_ENABLED']:
        ssl_options = {
            
            'keyfile': app.config['SSL_KEY_FILE'],
            'certfile': app.config['SSL_CERT_FILE']
        }
    else:
        ssl_options = {}

    if app.config['RTSTORE_ENABLED']:
        # run the realtime store service
        rtstore.run()

    set_frontend_config()

    if mode == 'local':
        socketio.run(app, host=host, port=5000, **ssl_options)
    elif mode == 'dev':
        socketio.run(app, host=host, port=5000, debug=True, **ssl_options)
    elif mode == 'deploy':
        socketio.run(app, host=host, **ssl_options)
    else:
        raise f'unrecognized mode {mode}'

@covfee_cli.command()
@click.option("--host", default='localhost', help="Specify the IP address to serve webpack dev server. Use for testing in a local network.")
def webpack(host):
    """
    Launches a webpack instance for use in dev mode
    """
    _, app = create_app('dev')

    with app.app_context():
        covfee_folder = CovfeeFolder(os.getcwd())
        if not covfee_folder.is_project():
            return print(Fore.RED+'Working directory is not a valid covfee project folder. '
                                'Did you run covfee-maker in the current folder?')

        covfee_folder.launch_webpack(host=host)

@covfee_cli.command()
def build():
    """
    Builds the covfee bundle into the project folder
    """
    _, app = create_app('dev')

    with app.app_context():
        covfee_folder = CovfeeFolder(os.getcwd())
        if not covfee_folder.is_project():
            raise Exception(
                'Working directory is not a valid covfee project folder.')

        covfee_folder.build()

@covfee_cli.command(name="start-deepstream")
def launch_deepstream():
    """
    Starts the daemon in charge of updating the true store (state).
    Required for shared tasks.
    """
    start_deepstream()


@covfee_cli.command()
@click.option("--dev", is_flag=True, help="Run covfee in dev mode")
@click.option("--deploy", is_flag=True, help="Run covfee in deployment mode")
@click.option('--no-browser', is_flag=True, help='Disables launching of the web browser.')
@click.option('--safe', is_flag=True, help='Enables login protection in local and dev mode (disabled by default).')
def start(dev, deploy, no_browser, safe):
    """
    Starts covfee in local mode by default. Use --deploy or --dev to start deployment (public) or development servers.
    """
    assert not (dev and deploy), '--dev and --deploy are mutually exclusive'
    mode = 'local'
    if dev: mode = 'dev'
    if deploy: mode = 'deploy'
    unsafe = False if mode == 'deploy' else (not safe)

    socketio, app = create_app(mode)  

    with app.app_context():
        covfee_folder = CovfeeFolder(os.getcwd())
        if not covfee_folder.is_project():
            return print(Fore.RED+'Working directory is not a valid covfee project folder. Did you run'
                                ' covfee maker in the current folder?')

        if not no_browser:
            covfee_folder.launch_in_browser(unsafe)

        app.config['UNSAFE_MODE_ON'] = unsafe
        start_covfee(socketio, app, mode)


@covfee_cli.command(name="open")
def open_covfee():
    """
    Opens covfee (admin panel) in a browser window
    """
    _, app = create_app('dev')

    with app.app_context():
        covfee_folder = CovfeeFolder(os.getcwd())
        if not covfee_folder.is_project():
            raise Exception('Working directory is not a valid covfee project folder.')
        covfee_folder.launch_in_browser()


def install_npm_packages(force=False):
    server_path = app.config['COVFEE_SERVER_PATH']
    npm_package = NPMPackage(server_path)
    if force or not npm_package.is_installed():
        npm_package.install()

    shared_path = app.config['COVFEE_SHARED_PATH']
    npm_package = NPMPackage(shared_path)
    if force or not npm_package.is_installed():
        npm_package.install()

@covfee_cli.command()
def installjs():
    _, app = create_app('local')
    with app.app_context():
        install_npm_packages(force=True)

@covfee_cli.command()
@click.option("--force", is_flag=True, help="Specify to overwrite existing databases.")
@click.option("--deploy", is_flag=True, help="Run covfee in deployment / production mode.")
@click.option("--safe", is_flag=True, help="Enable authentication in local mode.")
@click.option("--rms", is_flag=True, help="Re-makes the schemata for validation.")
@click.option("--no-launch", is_flag=True, help="Do not launch covfee, only make the DB")
@click.option("--no-browser", is_flag=True, help="Do not launch in the browser")
@click.argument("file_or_folder")
def make(force, deploy, safe, rms, no_browser, no_launch, file_or_folder):

    mode = 'deploy' if deploy else 'local'
    unsafe = False if deploy else (not safe)
    socketio, app = create_app(mode)

    with app.app_context():
        install_npm_packages()

        try:
            covfee_make(file_or_folder, force=force, rms=rms, stdout_enabled=True)
        except FileNotFoundError:
            pass
        except JavascriptError as err:
            return print('This is likely an issue with the Covfee app. Please contact the developers or post an issue with the following error message. \n' + err.js_stack_trace)
        except ValidationError as err:
            return err.print_friendly()
        except ProjectExistsException as err:
            return print('Project exists in covfee. Add --force option to overwrite.')
        except Exception as err:
            print(traceback.format_exc())
            if 'js_stack_trace' in dir(err):
                print(err.js_stack_trace)
            return

        if no_launch: 
            return

        if not no_browser:
            project_folder = CovfeeFolder(os.getcwd())
            project_folder.launch_in_browser(unsafe)

        app.config['UNSAFE_MODE_ON'] = unsafe
        start_covfee(socketio, app, 'local')

@covfee_cli.command()
def mkuser():
    """
    Creates an admin in the database.
    """
    _, app = create_app('dev')

    with app.app_context():
        project_folder = CovfeeFolder(os.getcwd())
        if not project_folder.is_project():
            project_folder.init()

        username = input('Please enter username: ')
        password = getpass()
        project_folder.mkuser(username, password)



import os
import hmac
import hashlib
import datetime
import random

from flask import current_app as app
from werkzeug.datastructures import MultiDict

from .db import db
from hashlib import sha256
from covfee.server.orm.task import TaskSpec

hits_taskspecs = db.Table(
    'hits_taskspecs',
    db.Column('hit_id', db.Integer, db.ForeignKey('hits.id'), primary_key=True),
    db.Column('taskspec_id', db.Integer, db.ForeignKey('taskspecs.id'), primary_key=True))


class HIT(db.Model):
    ''' Represents a set of tasks to be completed by one subject, ie. a HIT 
        - A HIT belongs to a project and has a list of Task specifications.
        - A HIT is abstract and must be instantiated into a HIT instance to be solved by a user. 
    '''
    __tablename__ = 'hits'
    __table_args__ = (
        db.UniqueConstraint('project_id', 'name'),
    )

    id = db.Column(db.LargeBinary, primary_key=True)
    project_id = db.Column(db.Integer, db.ForeignKey('projects.id'))
    # backref projec

    instances = db.relationship(
        "HITInstance", backref='hit', cascade="all, delete")
    taskspecs = db.relationship(
        "TaskSpec", secondary=hits_taskspecs, backref='hits', cascade="all, delete")

    # shared_tasks = db.relationship(
    #     "Task", backref='hit', cascade="all, delete")

    name = db.Column(db.String)
    extra = db.Column(db.JSON)
    interface = db.Column(db.JSON)
    config = db.Column(db.JSON)

    created_at = db.Column(db.DateTime, default=datetime.datetime.now)
    updated_at = db.Column(db.DateTime, onupdate=datetime.datetime.now)

    def __init__(self, id, project_id, name, interface={}, extra=None, tasks=[], config={}, **kwargs):
        self.id = sha256(f'{id}_{project_id}_{app.config["COVFEE_SECRET_KEY"]}'.encode()).digest()
        self.name = name

        self.interface = interface

        if extra is not None:
            if 'url' in extra and extra['url'][:4] != 'http':
                extra['url'] = os.path.join(app.config['PROJECT_WWW_URL'], extra['url'])
        self.extra = extra

        self.config = config
        self.config['maxInstances'] = self.config.get('maxInstances', 100)

        self.set_task_specs(tasks)

    def update(self, params):
        self.config = {**self.config, **params['config']}

    def set_task_specs(self, tasks_spec):
        ''' This method will separate the tasks specification into a generator specification
            that will specify how a HIT is instantiated (incl shuffling of the tasks) and refers to tasks via IDs
            and a list of tasks, containing only covfee tasks (excl shufflers), and whose index in the list match
            the IDs in the generator.
        '''
        
        task_list = []
        state = {"task_idx": 0}

        # traverse the tasks_spec looking for covfee tasks
        def dfs(ts):
            for i, spec in enumerate(ts):
                if spec['type'] == 'shuffle':
                    for child_ts in spec['tasks']:
                        dfs(child_ts)
                else:
                    task_list.append(spec)
                    ts[i] = state['task_idx']
                    state['task_idx'] += 1

        # recurse starting at the root HIT tasks spec
        dfs(tasks_spec)

        # store the covfee tasks specifications
        for i, spec in enumerate(task_list):
            if 'name' not in spec:
                spec['name'] = f'Task {i}'
            self.taskspecs.append(TaskSpec(**spec))
        
        # store the generator specification
        self.config['generator'] = tasks_spec

    def get_hashstr(self, id):
        ''' Used to generate the string to be hashed as an ID to the HIT instances of this HIT.
        '''
        return self.id.hex() + id

    def instantiate(self):
        ''' Generates a new hit instance/URL by instantiating the HIT.
            Task specs are arranged per config.generator and passed to TaskInstance in final order
        '''
        if(len(self.instances) >= self.config['maxInstances']):
            return None

        task_specs = []
        def recursive_generate(gen_list):
            ''' Recursive method that will DFS the generator structure.
                - new tasks are added to task_specs when an integer ID is found
                - the recursive step is taken on special objects (eg. type='shuffle')
            '''
            for elem in gen_list:
                if type(elem) == int:
                    task_specs.append(self.taskspecs[elem])
                elif type(elem) == dict:
                    if elem['type'] == 'shuffle':
                        idxs = random.sample(range(len(elem['tasks'])), len(elem['tasks']))
                        shuffled = [elem['tasks'][i] for i in idxs]
                        for child_gen_list in shuffled:
                            recursive_generate(child_gen_list)
                    else:
                        raise Exception(f'Unrecognized element type property "{elem["type"]}" in generator.')
                else:
                    raise Exception(f'Unrecognized element type "{str(type(elem))}" in generator.')

        # start recursion on the root generator
        recursive_generate(self.config['generator'])    

        instance = HITInstance(
            id=sha256(self.get_hashstr(f'instance{len(self.instances)}').encode()).digest(),
            submitted=False,
            taskspecs=task_specs
        )
        self.instances.append(instance)
        return instance

    def get_api_url(self):
        return f'{app.config["API_URL"]}/hits/{self.id.hex()}'

    def as_dict(self, with_project=True, with_instances=False, with_instance_tasks=False, with_config=False):
        hit_dict = {c.name: getattr(self, c.name)
                    for c in self.__table__.columns}
        hit_dict['id'] = hit_dict['id'].hex()
        hit_dict['api_url'] = self.get_api_url()
        hit_dict['generator_url'] = self.get_generator_url()

        if with_instances:
            hit_dict['instances'] = [instance.as_dict(
                with_tasks=with_instance_tasks) for instance in self.instances]

        if with_project:
            hit_dict['project'] = self.project.as_dict()
        del hit_dict['project_id']

        if not with_config:
            del hit_dict['config']

        return hit_dict
    
    def get_generator_url(self):
        ''' URL to the generator endpoint, which will instantiate the HIT and redirect the user to the new instance
        For use in linking from crowdsourcing websites (eg. Prolific)
        '''
        return f'{app.config["API_URL"]}/hits/{self.id.hex():s}/instances/add_and_redirect'


class HITInstance(db.Model):
    ''' Represents an instance of a HIT, to be solved by one user
        - one HIT instance maps to one URL that can be sent to a participant to access and solve the HIT.
        - a HIT instance is specified by the abstract HIT it is an instance of.
        - a HIT instance is linked to a list of tasks (instantiated task specifications),
        which hold the responses for the HIT
    '''
    __tablename__ = 'hitinstances'

    id = db.Column(db.LargeBinary, primary_key=True)
    # id used for visualization
    preview_id = db.Column(db.LargeBinary, unique=True)
    hit_id = db.Column(db.LargeBinary, db.ForeignKey('hits.id'))
    # backref hit

    tasks = db.relationship("Task", backref='hitinstance', cascade="all, delete")
    submitted = db.Column(db.Boolean)
    queryParams = db.Column(db.JSON)

    created_at = db.Column(db.DateTime, default=datetime.datetime.now)
    updated_at = db.Column(db.DateTime, onupdate=datetime.datetime.now)
    submitted_at = db.Column(db.DateTime)

    def __init__(self, id, taskspecs=[], submitted=False):
        self.id = id
        self.preview_id = sha256((id + 'preview'.encode())).digest()
        self.submitted = submitted
        self.queryParams = {}

        for spec in taskspecs:
            self.tasks.append(spec.instantiate())

    def get_api_url(self):
        return f'{app.config["API_URL"]}/instances/{self.id.hex():s}'

    def get_url(self):
        return f'{app.config["APP_URL"]}/hits/{self.id.hex():s}'

    def get_preview_url(self):
        return f'{app.config["APP_URL"]}/hits/{self.preview_id.hex():s}?preview=1'

    def get_completion_info(self):
        completion = {
            'completionCode': self.hit.config.get('completionCode', sha256((self.id.hex() + app.config['COVFEE_SECRET_KEY']).encode()).digest().hex()[:12]),
            'redirectName': self.hit.config.get('redirectName', None),
            'redirectUrl': self.hit.config.get('redirectUrl', None)
        }
        return completion

    def get_hmac(self):
        h = hmac.new(app.config['COVFEE_SECRET_KEY'].encode('utf-8'), self.id, hashlib.sha256 )
        return h.hexdigest()

    def set_extra(self, params: MultiDict):
        self.queryParams = {
            **self.queryParams,
            **params
        }

    def submit(self):
        if any([(task.spec.required and not task.has_valid_response()) for task in self.tasks]):
            # cant submit if at least one required task has no valid submissions
            return False, 'Some required tasks have no valid responses.'
        else:
            self.submitted = True
            self.submitted_at = datetime.datetime.now()
            return True, None

    def as_dict(self, with_tasks=False, with_response_info=False):
        instance_dict = {c.name: getattr(self, c.name) for c in self.__table__.columns}
        hit_dict = self.hit.as_dict()

        instance_dict['id'] = instance_dict['id'].hex()
        instance_dict['token'] = self.get_hmac()
        instance_dict['hit_id'] = instance_dict['hit_id'].hex()
        instance_dict['preview_id'] = instance_dict['preview_id'].hex()

        # merge hit and instance dicts
        instance_dict = {**hit_dict, **instance_dict}

        if with_tasks:
            prerequisite_tasks = [task for task in self.tasks if task.spec.prerequisite]
            prerequisites_completed = all([task.has_valid_response() for task in prerequisite_tasks])
            instance_dict['prerequisites_completed'] = prerequisites_completed
            if prerequisites_completed:
                instance_dict['tasks'] = [task.as_dict() for task in self.tasks]
            else:
                instance_dict['tasks'] = [task.as_dict() for task in prerequisite_tasks]

        if self.submitted:
            instance_dict['completionInfo'] = self.get_completion_info()

        return instance_dict

    def stream_download(self, z, base_path, csv=False):
        for i, task in enumerate(self.tasks):
            yield from task.stream_download(z, base_path, i, csv)




import datetime
import json
import os
import sys
from io import BytesIO
from typing import Any, Tuple

import numpy as np
from flask import current_app as app
from sqlalchemy.orm import backref

from .db import db
from .. import tasks
from ..tasks.base import BaseCovfeeTask
from ..utils.packer import Packer
pytype = type


class TaskSpec(db.Model):
    """ Represents a task specification.
        Task specifications include the user-provided params (spec) for the task. It can be divided into:
            - Task-specific specification (spec), defined on the task's Typescript interface.
            - Generic task options: required, prerrequisite, maxSubmissions, autoSubmit, timer. These apply to all tasks.
            - Task children (children), recursively defined (points to the same table).
     """
    __tablename__ = 'taskspecs'
    id = db.Column(db.Integer, primary_key=True)
    parent_id = db.Column(db.Integer, db.ForeignKey('taskspecs.id'))
    # backref parent

    children = db.relationship("TaskSpec", backref=backref('parent', remote_side=[id]),
                               cascade="all, delete")
    tasks = db.relationship("Task", backref="spec", cascade="all, delete")

    required = db.Column(db.Boolean)
    prerequisite = db.Column(db.Boolean)
    order = db.Column(db.Integer)
    editable = db.Column(db.Boolean)
    spec = db.Column(db.JSON)
    config = db.Column(db.JSON)

    created_at = db.Column(db.DateTime, default=datetime.datetime.now)
    updated_at = db.Column(db.DateTime, onupdate=datetime.datetime.now)

    def __init__(self, maxSubmissions=0, autoSubmit=False, timer=None,
                 editable=False, required=False, prerequisite=False, shared=False, **spec):
        self.editable = editable
        self.required = required
        self.prerequisite = prerequisite
        self.config = {
            'maxSubmissions': maxSubmissions,
            'autoSubmit': autoSubmit,
            'timer': timer,
            'shared': shared
        }

        if 'children' in spec:
            for child in spec['children']:
                child_spec = TaskSpec(**child)
                self.children.append(child_spec)
            del spec['children']

        self.spec = spec

    def instantiate(self):
        # return singleton task for shared tasks
        if self.config['shared'] and len(self.tasks) > 0:
            return self.tasks[0]

        task = Task()
        self.tasks.append(task)

        if(self.children):
            for child in self.children:
                task.children.append(child.instantiate())
        return task

    def as_dict(self):
        spec_dict = {c.name: getattr(self, c.name) for c in self.__table__.columns
                     if c not in ['responses']}

        spec_dict = {**spec_dict, **spec_dict['config']}
        del spec_dict['config']
        return spec_dict


class Task(db.Model):
    """ A Task is an instantiation of a TaskSpec, associated to a HitInstance.
        It represents a task to be solved within a hit instance.
    """
    __tablename__ = 'tasks'

    id = db.Column(db.Integer, primary_key=True)
    parent_id = db.Column(db.Integer, db.ForeignKey('tasks.id'))
    # backref parent
    hitinstance_id = db.Column(db.LargeBinary, db.ForeignKey('hitinstances.id'))
    # backref hitinstance
    taskspec_id = db.Column(db.Integer, db.ForeignKey('taskspecs.id'))
    # backref spec

    responses = db.relationship("TaskResponse", backref='task', cascade="all, delete-orphan")
    children = db.relationship("Task", backref=backref('parent', remote_side=[id]),
                               cascade="all, delete-orphan")

    # response status
    has_unsubmitted_response = db.Column(db.Boolean)
    created_at = db.Column(db.DateTime, default=datetime.datetime.now)
    updated_at = db.Column(db.DateTime, onupdate=datetime.datetime.now)

    def __init__(self):
        self.has_unsubmitted_response = False
        self.add_response() # add initial empty response

    def get_task_object(self):
        task_class = getattr(tasks, self.spec.spec['type'], BaseCovfeeTask)
        task_object = task_class(task=self)
        return task_object

    def as_dict(self):
        # merge task and spec dicts
        task_dict = {c.name: getattr(self, c.name) for c in self.__table__.columns
                     if c not in ['responses']}
        if task_dict['hitinstance_id']:   # child tasks may have no hitinstance_id
            task_dict['hitinstance_id'] = task_dict['hitinstance_id'].hex()
        spec_dict = self.spec.as_dict()

        task_dict = {**spec_dict, **task_dict}

        task_dict['responses'] = [response.as_dict() for response in self.responses]
        # task is valid if any response is valid
        task_dict['valid'] = self.has_valid_response()
        if self.children:
            task_dict['children'] = [child.as_dict() for child in self.children]
        else:
            task_dict['children'] = []
        task_dict['num_submissions'] = sum([1 if res.submitted else 0 for res in self.responses])
        task_dict['url'] = f'{app.config["API_URL"]}/tasks/{task_dict["id"]}'

        task_object = self.get_task_object()
        task_dict['taskSpecific'] = task_object.get_task_specific_props()

        return task_dict

    def has_valid_response(self):
        return any([response.valid for response in self.responses])

    def add_response(self):
        response = TaskResponse()
        self.responses.append(response)
        return response

    def stream_download(self, z, base_path, index, csv=False):
        responses = [resp for resp in self.responses if resp.submitted]

        for i, response in enumerate(responses):
            if csv:
                # write the CSV data
                df = response.get_dataframe()
                if df is not None:
                    stream = BytesIO()
                    df.to_csv(stream, mode='wb')
                    stream.seek(0)
                    z.write_iter(os.path.join(base_path, response.get_download_filename(task_index=index, response_index=i) + '.csv'),
                                stream)

            # write the json response
            response_dict = response.get_json(
                with_chunk_data=not csv)   # important
            stream = BytesIO()
            stream.write(json.dumps(response_dict).encode())
            stream.seek(0)
            z.write_iter(os.path.join(base_path, response.get_download_filename(task_index=index, response_index=i) + '.json'),
                         stream)

            yield from z.flush()
        
        for child in self.children:
            yield from child.stream_download(z, base_path, csv)


class TaskResponse(db.Model):
    """ Represents a task's response """
    __tablename__ = 'taskresponses'

    id = db.Column(db.Integer, primary_key=True)
    task_id = db.Column(db.Integer, db.ForeignKey('tasks.id'))
    # backref task

    chunks = db.relationship("Chunk", backref='taskresponse',
                             order_by="Chunk.index", cascade="all, delete-orphan", lazy="dynamic")

    state = db.Column(db.JSON) # holds the shared state of the task

    submitted = db.Column(db.Boolean)
    valid = db.Column(db.Boolean)
    data = db.Column(db.JSON)

    created_at = db.Column(db.DateTime, default=datetime.datetime.now)
    updated_at = db.Column(db.DateTime, onupdate=datetime.datetime.now)
    submitted_at = db.Column(db.DateTime)

    # can be used to store server state (eg. state of recording)
    extra = db.Column(db.JSON)

    def __init__(self):
        self.submitted = False
        self.valid = False
        self.extra = {}

    def as_dict(self):
        response_dict = {c.name: getattr(self, c.name) for c in self.__table__.columns}
        response_dict = {**response_dict}
        response_dict['url'] = f'{app.config["API_URL"]}/responses/{response_dict["id"]}'
        del response_dict['extra']
        
        return response_dict

    def get_task_object(self):

        task_class = getattr(tasks, self.task.spec.spec['type'], BaseCovfeeTask)
        task_object = task_class(response=self)
        return task_object

    def get_dataframe(self):
        task_object = self.get_task_object()
        chunk_data, chunk_logs = self.get_ndarray()
        df = task_object.to_dataframe(chunk_data)
        return df

    def get_json(self, with_chunk_data=True):
        task_object = self.get_task_object()
        # if with_chunk_data:
        #     chunk_data, chunk_logs = self.get_ndarray()
        #     task_json = task_object.to_dict(self.data, chunk_data)
        # else:
        #     task_json = task_object.to_dict(self.data, None)
        return task_object.to_dict(with_chunk_data)

    def get_download_filename(self, task_index, response_index):
        if self.task.parent:
            # start with the parent name for children tasks
            return f'{self.task.parent.spec.spec["name"]}-{self.task.spec.spec["name"]}_{response_index:d}'
        else:
            # use the task id if available
            if self.task.spec.spec.get('id', False):
                return f'{task_index}_{self.task.spec.spec["id"]}_{response_index:d}'
            return f'{task_index}_{self.task.spec.spec["name"]}_{response_index:d}'

    def pack_chunks(self):
        chunks = self.chunks.order_by(Chunk.index.desc()).all()
        chunk_bytes = len(chunks).to_bytes(4, sys.byteorder)
        chunk_bytes += b''.join([chunk.data for chunk in chunks])
        return chunk_bytes

    def get_ndarray(self) -> Tuple[np.ndarray, Any]:
        """This method takes care of aggregating a list of binary data chunks into a single numpy array.

        Args:
            data_chunks (list[bytes]): list of chunks as sent from the server. Size of the chunks may vary.

        Returns:
            np.ndarray: single numpy array with the aggregated data
        """
        if not self.chunks.count():
            return None, None

        chunks = [chunk.unpack() for chunk in self.chunks.all()]

        idxs_chunks = list()
        data_chunks = list()
        logs_chunks = list()

        for chunk in chunks:
            data = chunk['data']
            chunk_length = chunk['chunkLength']

            if len(data) % chunk_length != 0:
                raise Exception('Chunk byte length is invalid.')
            bytes_per_record = len(data) // chunk_length

            if (bytes_per_record - 4) % 8 != 0:
                raise Exception(
                    'Number of data bytes per record is not an 8-multiple.')
            datapoints_per_record = (bytes_per_record - 4) // 8
            idxs_chunks.append(np.frombuffer(
                data,
                dtype=np.uint32,
                count=chunk_length,
                offset=0).astype(np.float64).reshape(-1, 1))
            data_chunks.append(np.frombuffer(
                data,
                dtype=np.float64,
                count=chunk_length * datapoints_per_record,
                offset=4 * chunk_length).reshape(-1, datapoints_per_record))
            logs_chunks.append(chunk['logs'])

        idxs = np.vstack(idxs_chunks)
        data = np.vstack(data_chunks)
        assert len(idxs) == len(data)

        return np.hstack([idxs, data]), [l for chunk in logs_chunks for l in chunk]

    def validate(self):
        task_object = self.get_task_object()
        chunk_data, chunk_logs = self.get_ndarray()
        self.valid = task_object.validate(self.data, chunk_data, chunk_logs)
        return self.valid

    def submit(self, response=None):
        
        validation_result = self.validate()
        
        if response is not None:
            self.data = response
        self.submitted = True
        self.submitted_at = datetime.datetime.now()
        self.valid = (validation_result == True)
        self.task.has_unsubmitted_response = False

        res = {
            'status': 'success',
            'valid': self.valid,
            'response': self.as_dict()
        }

        if not self.valid:
            res['reason'] = validation_result

        return res

# represents a chunk of task response (for continuous responses)
class Chunk(db.Model):
    """ Represents a chunk of or partial task response"""
    __tablename__ = 'chunks'

    # (taskresponse, index) must be unique
    index = db.Column(db.Integer, primary_key=True)
    taskresponse_id = db.Column(db.Integer, db.ForeignKey(
        'taskresponses.id'), primary_key=True)
    # ini_time = db.Column(db.Float, index=True)
    # end_time = db.Column(db.Float, index=True)

    data = db.Column(db.LargeBinary)
    length = db.Column(db.Integer)   # number of samples in data
    log_data = db.Column(db.JSON, nullable=True)

    def __init__(self, index, data, length, log_data=None):
        self.index = index
        self.update(data, length, log_data)

    def update(self, data, length, log_data=None):
        self.data = data
        self.length = length
        self.log_data = log_data
        # self.ini_time = data[0][0]
        # self.end_time = data[-1][0]

    def as_dict(self):
        chunk_dict = {c.name: getattr(
            self, c.name) for c in self.__table__.columns}
        return chunk_dict

    def unpack(self):
        packer = Packer()
        parsed = packer.parseChunk(self.data)
        return parsed



import os
import json
from collections import Counter

from flask import current_app as app
from covfee.cli.utils import working_directory


class Schemata:
    def __init__(self):
        self.schemata = None

        self.schemata_path = app.config['SCHEMATA_PATH']

    def exists(self):
        return os.path.exists(self.schemata_path)

    def get(self):
        if self.schemata is not None:
            return self.schemata

        if self.exists():
            self.schemata = json.load(open(self.schemata_path))
            return self.schemata

        self.make()
        return self.schemata

    def recursive_resolve_refs(self, definition):
        def resolve(node):
            # Returns a list of the resolved children nodes of a node up to nodes with properties
            if 'allOf' in node:
                raise Exception('found allOf in schema.')

            if 'oneOf' in node:
                raise Exception('found oneOf in node')

            if '$ref' in node:
                return resolve(self.get_ref(node['$ref']))

            if 'anyOf' in node:
                node['anyOf'] = [resolve(n) for n in node['anyOf']]
                return node

            if node['type'] == 'object' and 'properties' in node and node['properties']:
                node['properties'] = {k: resolve(n) for k, n in node['properties'].items()}

            if node['type'] == 'array' and node['items']:
                if type(node['items']) == dict:
                    node['items'] = resolve(node['items'])
                else:
                    node['items'] = [resolve(n) for n in node['items']]

            return node

        try:
            return resolve(definition)
        except:
            pass

        

        

    def make(self):
        try:
            os.remove(app.config["SCHEMATA_PATH"])
        except OSError:
            pass
        # make the typescript into json schemata
        with working_directory(app.config['COVFEE_SHARED_PATH']):
            tsconfig_path = os.path.join(app.config['COVFEE_SHARED_PATH'], 'tsconfig.json')
            cmd = f'npx typescript-json-schema {tsconfig_path} "MyProjectSpec" --titles --ignoreErrors --required -o {app.config["SCHEMATA_PATH"]}'
            os.system(cmd)

        # process the schemata for validation
        schemata = json.load(open(app.config["SCHEMATA_PATH"]))
        defs_only = {
            '$schema': schemata['$schema'],
            'definitions': schemata['definitions']
        }
        self.schemata = defs_only

        self.schemata['definitions'] = {
            k: self.add_discriminators(d)
            for k, d in self.schemata['definitions'].items()}
        json.dump(self.schemata, open(
            app.config['SCHEMATA_PATH'], 'w'), indent=2)

    def get_ref(self, ref):
        return self.schemata['definitions'][ref[14:]]

    def add_discriminators(self, definition):

        def resolve(node):
            # Returns a list of the resolved children nodes of a node up to nodes with properties
            if 'allOf' in node:
                raise Exception('found allOf in schema.')

            if 'oneOf' in node:
                raise Exception('found oneOf in node')

            if '$ref' in node:
                return resolve(self.get_ref(node['$ref']))

            if 'anyOf' in node:
                return [e for n in node['anyOf'] for e in resolve(n)]

            return [node]

        # look for a const property with a default value
        def get_default_const_property(node):
            for k, n in node['properties'].items():
                if 'default' in n and 'enum' in n and len(n['enum']) == 1 and n['default'] == n['enum'][0]:
                    return k

            return False

        # returns a node with cases of anyOf with conditional property transformed into if then else statements.
        # $refs are resolved to avoid AJV bug: https://github.com/ajv-validator/ajv/pull/1815
        def recursive_dfs(node, path=[]):
            if '$ref' in node:
                return node

            if 'anyOf' in node:
                # resolve children
                children = resolve(node)
                # do children have default const properties?
                children_const_props = [get_default_const_property(c) for c in children]
                most_common, count = Counter(children_const_props).most_common(1)[0]

                if count != len(children_const_props):
                    raise 'found a default const element for some but not all of children nodes'

                if most_common is False:
                    node['anyOf'] = [recursive_dfs(n) for n in node['anyOf']]
                    return node

                pivot_prop = most_common
                # there is a default const property in all children
                # modify the schema accordingly
                for child in children:
                    child['additionalProperties'] = False
                node['oneOf'] = node['anyOf']
                del node['anyOf']
                node['discriminator'] = {'propertyName': pivot_prop}
                if 'required' not in node:
                    node['required'] = []
                node['required'].append(pivot_prop)

                return node

            if 'type' not in node:
                raise 'type not found in node'

            if node['type'] == 'object' and 'properties' in node and node['properties']:
                node['properties'] = {
                    k: recursive_dfs(n) for k, n in node['properties'].items()}

            if node['type'] == 'array' and node['items']:
                if type(node['items']) == dict:
                    node['items'] = recursive_dfs(node['items'])
                else:
                    node['items'] = [recursive_dfs(n) for n in node['items']]

            return node

        return recursive_dfs(definition)



import os
import sys
from collections import OrderedDict

from .access_token import *

Role_Attendee = 0 # depreated, same as publisher
Role_Publisher = 1 # for live broadcaster
Role_Subscriber = 2 # default, for live audience
Role_Admin = 101 # deprecated, same as publisher


class RtcTokenBuilder:
    # appID: The App ID issued to you by Agora. Apply for a new App ID from 
    #        Agora Dashboard if it is missing from your kit. See Get an App ID.
    # appCertificate:	Certificate of the application that you registered in 
    #                  the Agora Dashboard. See Get an App Certificate.
    # channelName:Unique channel name for the AgoraRTC session in the string format
    # uid: User ID. A 32-bit unsigned integer with a value ranging from 
    #      1 to (232-1). optionalUid must be unique.
    # role: Role_Publisher = 1: A broadcaster (host) in a live-broadcast profile.
    #       Role_Subscriber = 2: (Default) A audience in a live-broadcast profile.
    # privilegeExpireTs: represented by the number of seconds elapsed since 
    #                    1/1/1970. If, for example, you want to access the
    #                    Agora Service within 10 minutes after the token is 
    #                    generated, set expireTimestamp as the current 
    #                    timestamp + 600 (seconds)./
    @staticmethod
    def buildTokenWithUid(appId, appCertificate, channelName, uid, role, privilegeExpiredTs):
        return RtcTokenBuilder.buildTokenWithAccount(appId, appCertificate, channelName, uid, role, privilegeExpiredTs)

    # appID: The App ID issued to you by Agora. Apply for a new App ID from 
    #        Agora Dashboard if it is missing from your kit. See Get an App ID.
    # appCertificate:	Certificate of the application that you registered in 
    #                  the Agora Dashboard. See Get an App Certificate.
    # channelName:Unique channel name for the AgoraRTC session in the string format
    # userAccount: The user account. 
    # role: Role_Publisher = 1: A broadcaster (host) in a live-broadcast profile.
    #       Role_Subscriber = 2: (Default) A audience in a live-broadcast profile.
    # privilegeExpireTs: represented by the number of seconds elapsed since 
    #                    1/1/1970. If, for example, you want to access the
    #                    Agora Service within 10 minutes after the token is 
    #                    generated, set expireTimestamp as the current 
    @staticmethod
    def buildTokenWithAccount(appId, appCertificate, channelName, account, role, privilegeExpiredTs):
        token = AccessToken(appId, appCertificate, channelName, account)
        token.addPrivilege(kJoinChannel, privilegeExpiredTs)
        if (role == Role_Attendee) | (role == Role_Admin) | (role == Role_Publisher):
            token.addPrivilege(kPublishVideoStream, privilegeExpiredTs)
            token.addPrivilege(kPublishAudioStream, privilegeExpiredTs)
            token.addPrivilege(kPublishDataStream, privilegeExpiredTs)
        return token.build()


import pandas as pd
from .base import BaseCovfeeTask


class ContinuousKeypointTask(BaseCovfeeTask):

    def process_response(result, chunks, hitinstance, task):
        return {
            'hit_name': hitinstance.hit.name,
            'task_name': task.spec.name,
            'data': [x for y in chunks for x in y]
        }



import os
import hmac
import hashlib

from functools import wraps

from google.oauth2 import id_token
from google.auth.transport import requests
from flask import abort, request, jsonify, Blueprint, current_app as app
from flask_jwt_extended import (
    JWTManager, jwt_required, create_access_token, create_refresh_token, get_current_user,
    set_access_cookies, set_refresh_cookies, jwt_refresh_token_required, unset_jwt_cookies,
    get_jwt_identity, get_jwt_claims, verify_jwt_in_request
)

from ..orm import db, User, AuthProvider, password_hash

# AUTHENTICATION
# Using the user_claims_loader, we can specify a method that will be
# called when creating access tokens, and add these claims to the said
# token. This method is passed the identity of who the token is being
# created for, and must return data that is json serializable
# Here is a custom decorator that verifies the JWT is present in
# the request, as well as insuring that this user has a role of
# `admin` in the access token
def admin_required(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        if app.config.get('UNSAFE_MODE_ON', False):
            return fn(*args, **kwargs)
        verify_jwt_in_request()
        claims = get_jwt_claims()
        if 'admin' not in claims['roles']:
            return jsonify(msg='Admins only!'), 403
        else:
            return fn(*args, **kwargs)
    return wrapper

# This function is called whenever a protected endpoint is accessed,
# and must return an object based on the tokens identity.
# This is called after the token is verified, so you can use
# get_jwt_claims() in here if desired. Note that this needs to
# return None if the user could not be loaded for any reason,
# such as not being found in the underlying data store


def user_loader_callback(identity):
    return db.session.query(User).get(identity)
    # return User.query.filter_by(username=identity).first()

# Create a function that will be called whenever create_access_token
# is used. It will take whatever object is passed into the
# create_access_token method, and lets us define what the identity
# of the access token should be.


def user_identity_lookup(user):
    return int(user.id)

# Create a function that will be called whenever create_access_token
# is used. It will take whatever object is passed into the
# create_access_token method, and lets us define what custom claims
# should be added to the access token.


def add_claims_to_access_token(user):
    return {'roles': user.roles}


# AUTH routes
auth = Blueprint('auth', __name__)

def login_user(user):
    ''' Logs a user and returns their tokens
    '''

    # Create the tokens we will be sending back to the user
    access_token = create_access_token(identity=user)
    refresh_token = create_refresh_token(identity=user)

    res = jsonify({
        'id': user.id,
        'username': user.username,
        'roles': user.roles
    })
    set_access_cookies(res, access_token)
    set_refresh_cookies(res, refresh_token)
    return res, 200


@auth.route('/login-password', methods=['POST'])
def login_password():
    username = request.json.get('username', None)
    password = request.json.get('password', None)

    if username is None or password is None:
        return jsonify({"msg": "Bad username or password"}), 401

    user = User.query.filter_by(username=username).first()
    if user is None:
        return jsonify({"msg": "Bad username or password"}), 401

    password_providers = [prov for prov in user.providers if prov.provider_id == 'password']
    if len(password_providers) != 1:
        return jsonify({"msg": "Bad username or password"}), 401
    
    provider = password_providers[0]

    if provider.extra['password'] != password_hash(password).hex():
        return jsonify({"msg": "Bad username or password"}), 401

    return login_user(provider.user)


@auth.route('/login-google', methods=['POST'])
def login_google():
    token = request.json.get('token', None)

    if token is None:
        return jsonify({"msg": "Missing auth token"}), 401

    try:
        # Specify the CLIENT_ID of the app that accesses the backend:
        idinfo = id_token.verify_oauth2_token(token, requests.Request(), app.config.get('GOOGLE_CLIENT_ID', None))

        # ID token is valid. Get the user's Google Account ID from the decoded token.
        userid = idinfo['sub']
    except ValueError:
        # Invalid token
        return jsonify({"msg": "Invalid auth token"}), 401

    provider = AuthProvider.query.filter_by(
        user_id=userid, provider_id='google').first()

    if provider is None:
        # create the new user
        user = User(userid, roles=['user'])
        user.add_provider('google', userid)
        db.session.add(user)
        db.session.commit()
        return login_user(user)
    else:
        return login_user(provider.user)


@auth.route('/signup-password', methods=['POST'])
def signup():
    username = request.json.get('username', None)
    password = request.json.get('password', None)

    if username is None or password is None:
        return jsonify({"msg": "Missing username or password"}), 401

    provider = AuthProvider.query.filter_by(
        user_id=username, provider='password').first()

    if provider is not None:
        return jsonify({"msg": "Username exists"}), 401

    # Create the new user
    user = User(username, roles=['user'])
    user.add_provider('password', username, {'password': password})
    return login_user(user)



@auth.route('/refresh', methods=['POST'])
@jwt_refresh_token_required
def refresh():
    user = get_current_user()
    access_token = create_access_token(identity=user)
    res = jsonify({'username': user.username})
    set_access_cookies(res, access_token)
    return res, 200


@auth.route('/logout', methods=['POST'])
def logout():
    res = jsonify({'success': True})
    unset_jwt_cookies(res)
    return res, 200



# USER MANAGEMENT
@auth.route('/users/<uid>/delete')
@admin_required
def user_delete(uid):
    # Get the new user
    user = db.session.query(User).get(bytes.fromhex(uid))
    if user is None:
        return jsonify({'msg': 'invalid user'}), 400
    db.session.delete(user)
    db.session.commit()
    return jsonify({'success': True}), 200


import os
import shutil
import sys
import subprocess
import random
import traceback
import platform

from flask import current_app as app
from halo import Halo
from colorama import init as colorama_init, Fore

from covfee.server.orm import db
from covfee.server.orm.user import User, password_hash
from covfee.cli.utils import working_directory
from pathlib import Path
from covfee.shared.validator.ajv_validator import AjvValidator
from covfee.shared.validator.validation_errors import ValidationError
import json
from covfee.server.orm.project import Project
from shutil import which
colorama_init()


def cli_create_tables():
    '''
    Creates all the tables defined in the ORM
    '''
    with Halo(text='Creating tables', spinner='dots') as spinner:
        db.create_all()
        spinner.succeed('Created database tables.')


class ProjectExistsException(Exception):
    pass


class CovfeeFolder:

    def __init__(self, path):
        if not os.path.isdir(path):
            raise FileNotFoundError('covfee folder must be a valid folder.')
        self.path = path
        self.covfee_files = []
        self.projects = []

    def is_project(self):
        return os.path.exists(app.config['DATABASE_PATH'])

    def clear(self):
        shutil.rmtree(os.path.join(self.path, '.covfee'))

    def init(self):
        covfee_hidden = os.path.join(self.path, '.covfee')
        if not os.path.exists(covfee_hidden):
            os.makedirs(covfee_hidden)
        media_path = os.path.join(self.path, 'www', 'media')
        if not os.path.exists(media_path):
            os.makedirs(media_path)
        cli_create_tables()

    def add_covfee_files(self, file_or_folder):
        self.covfee_files = []

        # database exists and tables are created
        if os.path.isdir(file_or_folder):
            for json_path in Path(file_or_folder).rglob('*.covfee.json'):
                self.covfee_files.append(json_path)
        elif os.path.isfile(file_or_folder):
            self.covfee_files.append(file_or_folder)

        return self.covfee_files

    def validate_project(self, project_spec, filename, with_spinner=False):
        filter = AjvValidator()
        filter.validate_project(project_spec)

    def validate(self, with_spinner=False):
        for cf in self.covfee_files:
            # load the json file
            with Halo(text=f'Reading file {cf} as json..',
                      spinner='dots',
                      enabled=with_spinner) as spinner:
                try:
                    project_spec = json.load(open(cf))
                except Exception as e:
                    spinner.fail(f'Error reading file {cf} as JSON. Are you sure it is valid json?')
                    raise e
                spinner.succeed(f'Read covfee file {cf}.')

            with Halo(text=f'Validating project {project_spec["name"]} in {cf}',
                      spinner='dots',
                      enabled=with_spinner) as spinner:

                try:
                    self.validate_project(project_spec, cf, with_spinner=with_spinner)
                except Exception as e:
                    spinner.fail(f'Error validating project \"{project_spec["name"]}\" in {cf}.\n')
                    raise e
                self.projects.append((project_spec, cf))
                spinner.succeed(f'Project \"{project_spec["name"]}\" in {cf} is valid.')

    def push_projects(self, force=False, with_spinner=False):
        for project_spec, cf in self.projects:
            # delete existing project with the same id
            with Halo(text=f'Looking for existing projects with id {project_spec["id"]}',
                      spinner='dots',
                      enabled=False) as spinner:

                projid = Project.get_id(project_spec['id'])
                existing_project = Project.query.filter_by(id=projid).first()
                if existing_project is not None:
                    if force:
                        db.session.delete(existing_project)
                        db.session.commit()
                        spinner.warn(
                            f'Deleted existing project with conflicting ID {project_spec["id"]}')
        
                    else:
                        msg = f'Project with ID {project_spec["id"]} already exists.'
                        spinner.fail(msg)
                        raise ProjectExistsException(msg)

            # making
            with Halo(text=f'Making project {project_spec["name"]} from file {cf}',
                      spinner='dots',
                      enabled=False) as spinner:
                project = Project(**project_spec)
                db.session.add(project)
                spinner.succeed(
                    f'Project {project_spec["name"]} created successfully from file {cf}')
        db.session.commit()

    def link_bundles(self):
        master_bundle_path = os.path.join(app.config['MASTER_BUNDLE_PATH'], 'main.js')
        if not os.path.exists(master_bundle_path):
            raise Exception('Master bundles not found.')
        bundle_path = os.path.join(app.config['PROJECT_WWW_PATH'], 'main.js')
        if os.path.exists(bundle_path):
            os.remove(bundle_path)
        # windows requires admin rights for symlinking -> fall back to copying
        if(platform.system == 'Windows'):
            shutil.copyfile(
                master_bundle_path,
                bundle_path
            )
        else:
            os.symlink(
                master_bundle_path,
                bundle_path
            )

        admin_bundle_path = os.path.join(app.config['PROJECT_WWW_PATH'], 'admin.js')
        if os.path.exists(admin_bundle_path):
            os.remove(admin_bundle_path)
        if(platform.system == 'Windows'):
            shutil.copyfile(
                os.path.join(app.config['MASTER_BUNDLE_PATH'], 'admin.js'),
                admin_bundle_path
            )
        else:
            os.symlink(
                os.path.join(app.config['MASTER_BUNDLE_PATH'], 'admin.js'),
                admin_bundle_path
            )

    def launch_webpack(self, host=None):
        cwd = os.getcwd()
        # run the dev server
        with working_directory(app.config['COVFEE_CLIENT_PATH']):
            os.system('npx webpack serve' +
                      ' --env COVFEE_WD=' + cwd +
                      ' --config ./webpack.dev.js' +
                      ('' if host is None else ' --host ' + host))


    def launch_in_browser(self, unsafe=False):
        target_url = app.config["ADMIN_URL"] if unsafe else app.config["LOGIN_URL"]
        if which('xdg-open') is not None:
            os.system(f'xdg-open {target_url}')
        elif sys.platform == 'darwin' and which('open') is not None:
            os.system(f'open {target_url}')
        else:
            print(Fore.GREEN +
                  f' * covfee is available at {target_url}')

    def start_deepstream(self):
        # compile the TS to Js
        with working_directory(os.path.join(app.config['COVFEE_SERVER_PATH'], 'deepstream')):
            
            # generate a random password for deepstream
            password = random.getrandbits(128).to_bytes(16, 'little').hex()
            users_json = {   
                "admin": {
                    "password": password,
                    "serverData": {
                        "role": "superadmin"
                    }
                }
            }
            json.dump(users_json, open(os.path.join('..', 'conf', 'users.json'), 'w'))

            os.system('npx tsc')
            subprocess.Popen(['npx', 'deepstream', 'daemon'])
            subprocess.Popen(['npx', 'pm2', 'start', 'server.js', '-i', '1', '--watch', '--', 
                            'serve', password, 
                            str(app.config['DS_SERVER_PORT']), 
                            str(app.config['DS_CLIENT_PUB_PORT']), 
                            str(app.config['DS_CLIENT_SUB_PORT'])])


    def mkuser(self, username, password):
        user = User(username, ['admin'])
        user.add_provider('password', username, {'password': password_hash(password).hex()})
        db.session.add(user)
        db.session.commit()



from io import BytesIO

import zipstream
from flask import request, jsonify, Blueprint, send_file, make_response, Response,\
     stream_with_context, redirect
from sqlalchemy.orm.attributes import flag_modified

from ..orm import db, Project, HIT, HITInstance, Task, TaskResponse, Chunk
from .auth import admin_required
from covfee.server.orm.task import TaskSpec
api = Blueprint('api', __name__)


# PROJECTS
def jsonify_or_404(res, **kwargs):
    if res is None:
        return {'msg': 'not found'}, 404
    else:
        return jsonify(res.as_dict(**kwargs))


# return all projects
@api.route('/projects')
@admin_required
def projects():
    """Lists all the projects currently in covfee

    Returns:
        [type]: list of project objects
    """
    with_hits = request.args.get('with_hits', False)
    res = db.session.query(Project).all()
    if res is None:
        return jsonify([])
    else:
        return jsonify([p.as_dict(with_hits=with_hits) for p in res])


# return one project
@api.route('/projects/<pid>')
@admin_required
def project(pid):
    """Returns a project object

    Args:
        pid (str): project ID
    """
    with_hits = request.args.get('with_hits', False)
    with_instances = request.args.get('with_instances', False)
    res = db.session.query(Project).get(bytes.fromhex(pid))
    return jsonify_or_404(res, with_hits=with_hits, with_instances=with_instances, with_config=True)


@api.route('/projects/<pid>/csv')
@admin_required
def project_csv(pid):
    """Creates a CSV file with links and completion codes for HITs
    This file can be used in human intelligence marketplaces or to directly send the
    links in it to study participants.

    Args:
        pid ([str]): ID of the project

    Returns:
        [type]: CSV file with a hit instance per line
    """
    project = db.session.query(Project).get(bytes.fromhex(pid))
    if project is None:
        return {'msg': 'not found'}, 404
    else:
        df = project.get_dataframe()
        res = make_response(df.to_csv())
        res.headers["Content-Disposition"] = "attachment; filename=export.csv"
        res.headers["Content-Type"] = "text/csv"
        return res


@api.route('/projects/<pid>/download')
@admin_required
def project_download(pid):
    """Generates a downloadable with all the responses in a project.
    It gathers the responses for all hit instances in the project
    This endpoint migh be slow for large projects.

    Args:
        pid (str): project ID

    Returns:
        [type]: stream response with a compressed archive. 204 if the project has no responses
    """
    is_csv = bool(request.args.get('csv', False))

    project = db.session.query(Project).get(bytes.fromhex(pid))
    if project is None:
        return {'msg': 'not found'}, 404

    def generator():
        z = zipstream.ZipFile(mode='w', compression=zipstream.ZIP_DEFLATED)
        for chunk in project.stream_download(z, './', csv=is_csv):
            yield chunk
        yield from z

    response = Response(stream_with_context(generator()), mimetype='application/zip')
    response.headers['Content-Disposition'] = 'attachment; filename={}'.format('results.zip')
    return response


# HITS
# return one hit
@api.route('/hits/<hid>')
def hit(hid):
    """Returns a HIT specification.

    Args:
        hid (str): hit ID
    """
    with_instances = request.args.get('with_instances', False)
    with_instance_tasks = request.args.get('with_instance_tasks', False)
    res = db.session.query(HIT).get(bytes.fromhex(hid))
    return jsonify_or_404(res,
                          with_instances=with_instances,
                          with_instance_tasks=with_instance_tasks)

@api.route('/hits/<hid>/edit', methods=['POST'])
@admin_required
def hit_edit(hid):
    """ Edits the hit configuration using the received config.
    """
    hit = db.session.query(HIT).get(bytes.fromhex(hid))
    if hit is None:
        return jsonify({'msg': 'invalid hit'}), 400

    hit.update(request.json)
    db.session.commit()
    return jsonify_or_404(hit, with_instances=False, with_config=True)

# INSTANCES

# return one HIT instance
@api.route('/instances/<iid>')
def instance(iid):
    with_tasks = request.args.get('with_tasks', True)
    with_response_info = request.args.get('with_response_info', True)
    res = db.session.query(HITInstance).get(bytes.fromhex(iid))
    return jsonify_or_404(res, with_tasks=with_tasks, 
                          with_response_info=with_response_info)


@api.route('/instance-previews/<iid>')
def instance_preview(iid):
    res = HITInstance.query.filter_by(preview_id=bytes.fromhex(iid)).first()
    return jsonify_or_404(res, with_tasks=True, with_response_info=False)


# submit a hit (when finished)
@api.route('/instances/<iid>/submit', methods=['POST'])
def instance_submit(iid):
    instance = db.session.query(HITInstance).get(bytes.fromhex(iid))
    if instance is None:
        return jsonify({'msg': 'invalid instance'}), 400
    is_submitted, err = instance.submit()

    if not is_submitted:
        return jsonify({'msg': err}), 400

    db.session.commit()
    return jsonify(instance.as_dict(with_tasks=False))


@api.route('/hits/<hid>/instances/add')
def instance_add(hid):
    """Adds an instance (link) to the HIT

    Args:
        hid (str): HIT ID

    Returns:
        json: the added instance object
    """
    num_instances = request.args.get('num_instances', 1)
    hit = db.session.query(HIT).get(bytes.fromhex(hid))
    if hit is None:
        return jsonify({'msg': 'not found'}), 404

    for i in range(0, num_instances):
        hit.instantiate()

    db.session.commit()
    with_instances = request.args.get('with_instances', True)
    with_instance_tasks = request.args.get('with_instance_tasks', False)

    return jsonify_or_404(hit,
                          with_instances=with_instances,
                          with_instance_tasks=with_instance_tasks)


@api.route('/hits/<hid>/instances/add_and_redirect')
def instance_add_and_redirect(hid):
    """Adds an instance (link) to the HIT and redirects to the new URL.

    Args:
        hid (str): HIT ID
    """
    hit = db.session.query(HIT).get(bytes.fromhex(hid))
    if hit is None:
        return jsonify({'msg': 'not found'}), 404

    instance = hit.instantiate()
    if instance is None:
        return jsonify({'msg': 'unable to add instance.'}), 400
    instance.set_extra(request.args)

    db.session.commit()
    return redirect(instance.get_url(), 302)


@api.route('/instances/<iid>/download')
@admin_required
def instance_download(iid):
    """Generates a downloadable file with the results for a task instance
    It bundles together the results of individual tasks into a zip archive

    Args:
        iid ([type]): instance id

    Returns:
        [type]: zip file with task results
    """
    is_csv = request.args.get('csv', False)

    instance = db.session.query(HITInstance).get(bytes.fromhex(iid))
    if instance is None:
        return jsonify({'msg': 'not found'}), 404

    def generator():
        z = zipstream.ZipFile(mode='w', compression=zipstream.ZIP_DEFLATED)
        for chunk in instance.stream_download(z, './', csv=is_csv):
            yield chunk
        yield from z

    response = Response(stream_with_context(generator()), mimetype='application/zip')
    response.headers['Content-Disposition'] = 'attachment; filename={}'.format('results.zip')
    return response


# TASKS

# create a task attached to an instance
@api.route('/instances/<iid>/tasks/add', methods=['POST'])
def task_add_to_instance(iid):
    instance = db.session.query(HITInstance).get(bytes.fromhex(iid))
    if instance is None:
        return jsonify({'msg': 'invalid instance'}), 400

    task_spec = TaskSpec(**request.json, editable=True)
    task = task_spec.instantiate()
    instance.tasks.append(task)
    db.session.commit()
    return jsonify(task.as_dict())


# edit an existing task
@api.route('/tasks/<kid>/edit', methods=['POST'])
def task_edit(kid):
    task = db.session.query(Task).get(int(kid))
    if task is None:
        return jsonify({'msg': 'invalid task'}), 400
    spec = task.spec
    if not spec.editable:
        return jsonify(msg='Task cannot be user-edited.'), 403
    spec.spec['name'] = request.json['name']
    flag_modified(spec, 'spec')
    db.session.commit()
    return jsonify(task.as_dict())


# delete an existing task
@api.route('/tasks/<kid>/delete')
def task_delete(kid):
    task = db.session.query(Task).get(int(kid))
    if task is None:
        return jsonify({'msg': 'invalid task'}), 400
    if not task.editable:
        return jsonify(msg='Task cannot be user-deleted.'), 403
    db.session.delete(task)
    db.session.commit()
    return jsonify({'success': True}), 200


@api.route('/tasks/<kid>/response')
def response(kid):
    ''' Will return the last response for a task
    '''
    task = db.session.query(Task).get(int(kid))
    if task is None:
        return jsonify({'msg': 'invalid task'}), 400

    responses = task.responses
    submitted = request.args.get('submitted', None)
    if submitted is not None:
        responses = [r for r in responses if r.submitted == bool(submitted)]

    if len(responses) == 0:
        return jsonify(msg='No submitted responses found.'), 403

    response_dict = responses[-1].as_dict()
    return jsonify(response_dict)





@api.route('/tasks/<kid>/make_response', methods=['POST'])
def make_response(kid):
    submit = bool(request.args.get('submit', False))

    task = db.session.query(Task).get(int(kid))
    if task is None:
        return jsonify({'msg': 'invalid task'}), 400

    response = task.add_response()
    if request.json:
        response.update(request.json)
    if submit:
        response.submit()
    db.session.commit()
    return jsonify(response.as_dict())


# record a response to a task
@api.route('/responses/<rid>/submit', methods=['POST'])
def response_submit(rid):
    response = db.session.query(TaskResponse).get(int(rid))
    if response is None:
        return jsonify({'msg': 'invalid response'}), 400

    res = response.submit(request.json)    
    db.session.commit()
    return jsonify(res)

@api.route('/responses/<rid>/chunks', methods=['GET'])
def query_chunks(rid):
    response = db.session.query(TaskResponse).get(int(rid))
    if response is None:
        return jsonify({'msg': 'invalid response'}), 400

    chunk_bytes = response.pack_chunks()
    return send_file(BytesIO(chunk_bytes), mimetype='application/octet-stream'), 200

# receive a chunk of a response, for continuous responses
@api.route('/responses/<rid>/chunk', methods=['POST'])
def response_chunk(rid):
    sent_index = int(request.args.get('index'))
    length = int(request.args.get('length'))

    response = db.session.query(TaskResponse).get(int(rid))
    if response is None:
        return jsonify({'msg': 'invalid response'}), 400
    if response.submitted:
        return jsonify({'msg': 'response is already submitted'}), 400

    # if there is a previous chunk with the same index, overwrite it
    if response.chunks.count() > 0:
        chunk = next((chunk for chunk in response.chunks if chunk.index == sent_index), None)
        if chunk is not None:
            chunk.update(data=request.get_data(), length=length)

            db.session.commit()
            return jsonify({'success': True}), 201

    # no previous chunk with the same index -> append the chunk
    chunk = Chunk(index=sent_index, length=length, data=request.get_data())
    response.chunks.append(chunk)
    db.session.add(response)
    db.session.commit()
    return jsonify({'success': True}), 201




import versioneer
from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='covfee',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Jose Vargas",
    author_email="josedvq@gmail.com",
    description="Continuous video feedback tool",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    entry_points={
        'console_scripts': [
            # user CLI
            'covfee = covfee.cli.commands.launch:covfee_cli',
            # developer CLI
            'covfee-dev = covfee.cli.commands.dev:covfee_dev_cli'
        ]
    },
    install_requires=[
        'Flask == 1.*',
        'requests == 2.*',
        'flask_cors == 3.*',
        'Flask-SQLAlchemy == 2.*',
        'flask-socketio == 5.*',
        'gunicorn == 20.*',
        'flask-jwt-extended == 3.*',
        'click ==  7.*',
        'pandas == 1.*',
        'jsonschema == 3.*',
        'halo == 0.*',
        'colorama == 0.4.*',
        'pyzmq == 22.*',
        'numpy == 1.*',
        'zipstream-new == 1.*',
        'construct == 2.*',
        'eventlet == 0.*',
        'google-auth == 2.*'
    ],
    python_requires='>=3.6'
)



import json
import collections
from colorama import init, Fore
init()

class JavascriptError(Exception):
    pass

class ValidationError(Exception):
    '''Tracks and reports JSON schema validation errors.'''

    def __init__(self, message, path=[], instance=None):

        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        self.path = collections.deque()
        if type(path) == collections.deque:
            while True:
                try:
                    self.path.append(path.pop())
                except IndexError:
                    break
        elif type(path) == list:
            for idx in reversed(path):
                self.path.append(idx)

        self.instance = instance

    def append_path(self, *path_segments):
        for seg in reversed(path_segments):
            self.path.append(seg)

    def pop_path(self):
        self.path.pop()

    def get_python_path_string(self):
        '''Returns a string in python object access notation (using []) to the object where the error was generated.'''
        path_string = ''
        while True:
            try:
                path_segment = self.path.pop()
                if path_segment.isnumeric():
                    path_string += f'[{path_segment}]'
                else:
                    path_string += f'["{path_segment}"]'
            except IndexError:
                return path_string

    def prune_pprint(self, data, indent=4, threshold=10):

        def get_pruned(dt):
            if dt == dict:
                return {'$$pruned$$': '$$pruned$$'}, 3
            if dt == list:
                return ['$$pruned$$'], 3
            return '$$pruned$$', 1

        def prune_node(node):
            # get children display len
            if type(node) == dict:
                # approx number of lines that the node takes when printed
                children_display_lengths = []#[prune_node(n) for k,n in node.items()]
                for k,n in node.items():
                    child_len = prune_node(n)
                    if child_len > threshold:
                        node[k], child_len = get_pruned(type(n))
                    children_display_lengths.append(child_len)
                return 2 + sum(children_display_lengths)

            if type(node) == list:
                children_display_lengths = []
                for i, n in enumerate(node):
                    child_len = prune_node(n)
                    if child_len > threshold:
                        node[i], child_len = get_pruned(type(n))
                    children_display_lengths.append(child_len)
                return 2 + sum(children_display_lengths)

            if type(node) in [str, bool, float, int] or node is None:
                return max(1, len(str(node)) / 80)

        prune_node(data)
        printed = json.dumps(data, indent=indent).replace('"$$pruned$$": "$$pruned$$"', '...').replace('"$$pruned$$"', '...')

        return printed



    def print_friendly(self):
        print(f'\nError in project{self.get_python_path_string()} for object: ')
        print(Fore.BLUE+self.prune_pprint(self.instance, indent=4))
        print('Error: '+ Fore.RED+str(self))


import os

from halo.halo import Halo

from flask import current_app as app
from covfee.server.start import create_app
from covfee.shared.schemata import Schemata
from covfee.shared.validator.validation_errors import ValidationError
from .covfee_folder import CovfeeFolder

def _make(file_or_folder, force=False, rms=False, stdout_enabled=True):
    
    project_folder = CovfeeFolder(os.getcwd())

    # add the covfee files to the project
    with Halo(text='Adding .covfee.json files', spinner='dots', enabled=stdout_enabled) as spinner:
        covfee_files = project_folder.add_covfee_files(file_or_folder)

        if len(covfee_files) == 0:
            err = f'No valid covfee files found. Make sure that {file_or_folder} points to a file or to a folder containing .covfee.json files.'
            spinner.fail(err)
            raise FileNotFoundError(err)

        spinner.succeed(f'{len(covfee_files)} covfee project files found.')

    # validate the covfee files
    schema = Schemata()
    if rms or not schema.exists():
        schema.make()
    
    project_folder.validate(with_spinner=stdout_enabled)

    # init project folder if necessary
    if not project_folder.is_project():
        project_folder.init()
    project_folder.push_projects(force=force, with_spinner=stdout_enabled)

    # link bundles
    with Halo(text='Linking covfee bundles', spinner='dots', enabled=stdout_enabled) as spinner:
        try:
            project_folder.link_bundles()
        except Exception as e:
            spinner.fail('Error linking bundles. Aborted.')
            raise e
        spinner.succeed('covfee bundles linked.')


def make(*args, **kwargs):
    _, app = create_app('local')
    with app.app_context():
        return _make(*args, **kwargs, stdout_enabled=False)

def start_deepstream():
    _, app = create_app('local')

    with app.app_context():
        covfee_folder = CovfeeFolder(os.getcwd())
        if not covfee_folder.is_project():
            raise Exception(
                'Working directory is not a valid covfee project folder.')
                
        covfee_folder.start_deepstream()
from . import _version
__version__ = _version.get_versions()['version']
